#  Copyright © 2023 Ingram Micro Inc. All rights reserved.

from queue import Full, PriorityQueue

from django.utils import timezone


class DelayMessage:
    """Delay message.

    :param delivery_tag: The server-assigned and channel-specific delivery tag.
    :type delivery_tag: int
    :param payload: Transport payload.
    :type payload: dj_cqrs.dataclasses.TransportPayload
    :param eta: Time after which the message should be requeued.
    :type eta: datetime.datetime
    """

    def __init__(self, delivery_tag, payload, eta):
        self.delivery_tag = delivery_tag
        self.payload = payload
        self.eta = eta


class DelayQueue:
    """Queue for delay messages."""

    def __init__(self, max_size=None):
        if max_size is not None:
            assert max_size > 0, "Delay queue max_size should be positive integer."

        self._max_size = max_size
        self._queue = PriorityQueue()

    def get(self):
        """
        :rtype: DelayMessage
        """
        *_, delay_message = self._queue.get()
        return delay_message

    def get_ready(self):
        """Returns messages with expired ETA.

        :return: delayed messages generator
        :rtype: typing.Generator[DelayMessage]
        """
        while self.qsize():
            delay_message = self.get()
            if delay_message.eta > timezone.now():
                # Queue is ordered by message ETA.
                # Remaining messages should wait longer, we don't check them.
                self.put(delay_message)
                break

            yield delay_message

    def put(self, delay_message):
        """Adds message to queue.

        :param delay_message: DelayMessage instance.
        :type delay_message: DelayMessage
        """
        assert isinstance(delay_message, DelayMessage)
        if self.full():
            raise Full("Delay queue is full")

        self._queue.put((
            delay_message.eta.timestamp(),
            delay_message.delivery_tag,
            delay_message,
        ))

    def qsize(self):
        return self._queue.qsize()

    def full(self):
        return (
            self._max_size is not None
            and self.qsize() >= self._max_size
        )



#  Copyright © 2023 Ingram Micro Inc. All rights reserved.


class BaseTransport:
    """
    CQRS pattern can be implemented over any transport (AMQP, HTTP, etc.)
    All transports need to inherit from this base class.
    Transport must be set in Django settings:

    ``` py3

        CQRS = {
            'transport': 'dj_cqrs.transport.rabbit_mq.RabbitMQTransport',
        }
    ```
    """

    consumers = {}

    @staticmethod
    def produce(payload):
        """
        Send data from master model to replicas.

        Args:
            payload (dj_cqrs.dataclasses.TransportPayload): Transport payload from master model.
        """
        raise NotImplementedError

    @staticmethod
    def consume(*args, **kwargs):
        """Receive data from master model."""
        raise NotImplementedError

    @staticmethod
    def clean_connection(*args, **kwargs):
        """Clean transport connection. Here you can close all connections that you have"""
        raise NotImplementedError



#  Copyright © 2023 Ingram Micro Inc. All rights reserved.

import sys

from django.core.management.base import BaseCommand, CommandError

from dj_cqrs.constants import NO_QUEUE
from dj_cqrs.management.commands.cqrs_sync import (
    DEFAULT_BATCH,
    DEFAULT_PROGRESS,
    Command as SyncCommand,
)
from dj_cqrs.registries import MasterRegistry


class Command(BaseCommand):
    help = 'Diff synchronizer from CQRS replica stream.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--batch', '-b',
            help='Batch size',
            type=int,
            default=DEFAULT_BATCH,
        )
        parser.add_argument(
            '--progress', '-p',
            help='Display progress',
            action='store_true',
        )

    def handle(self, *args, **options):
        progress = self._get_progress(options)
        batch_size = self._get_batch_size(options)

        with sys.stdin as f:
            first_line = f.readline().strip()
            model = self._get_model(first_line)
            queue = self._get_queue(first_line)

            for pks_line in f:
                sync_kwargs = {
                    'cqrs_id': model.CQRS_ID,
                    'filter': '{{"id__in": {0}}}'.format(pks_line.strip()),
                    'progress': progress,
                    'batch': batch_size,
                }
                if queue:
                    sync_kwargs['queue'] = queue

                SyncCommand().handle(**sync_kwargs)

    @staticmethod
    def _get_model(first_line):
        cqrs_id = first_line.split(',')[0]
        model = MasterRegistry.get_model_by_cqrs_id(cqrs_id)

        if not model:
            raise CommandError('Wrong CQRS ID: {0}!'.format(cqrs_id))

        return model

    @staticmethod
    def _get_queue(first_line):
        queue = first_line.split(',')[-1]
        if queue != NO_QUEUE:
            return queue

    @staticmethod
    def _get_batch_size(options):
        return options.get('batch', DEFAULT_BATCH)

    @staticmethod
    def _get_progress(options):
        return bool(options.get('progress', DEFAULT_PROGRESS))



#  Copyright © 2022 Ingram Micro Inc. All rights reserved.
import logging
import signal
import threading
from pathlib import Path

from django.core.management.base import BaseCommand, CommandError
from watchfiles import watch
from watchfiles.filters import PythonFilter
from watchfiles.run import start_process

from dj_cqrs.registries import ReplicaRegistry


logger = logging.getLogger('django-cqrs')


def consume(**kwargs):
    import django
    django.setup()

    from dj_cqrs.transport import current_transport
    try:
        current_transport.consume(**kwargs)
    except KeyboardInterrupt:
        pass


def _display_path(path):
    try:
        return f'"{path.relative_to(Path.cwd())}"'
    except ValueError:  # pragma: no cover
        return f'"{path}"'


class WorkersManager:

    def __init__(
            self,
            consume_kwargs,
            workers=1,
            reload=False,
            ignore_paths=None,
            sigint_timeout=5,
            sigkill_timeout=1,
    ):
        self.pool = []
        self.workers = workers
        self.reload = reload
        self.consume_kwargs = consume_kwargs
        self.stop_event = threading.Event()
        self.sigint_timeout = sigint_timeout
        self.sigkill_timeout = sigkill_timeout

        if self.reload:
            self.watch_filter = PythonFilter(ignore_paths=ignore_paths)
            self.watcher = watch(
                Path.cwd(),
                watch_filter=self.watch_filter,
                stop_event=self.stop_event,
                yield_on_timeout=True,
            )

    def handle_signal(self, *args, **kwargs):
        self.stop_event.set()

    def run(self):
        for sig in [signal.SIGINT, signal.SIGTERM]:
            signal.signal(sig, self.handle_signal)
        if self.reload:
            signal.signal(signal.SIGHUP, self.restart)

        self.start()

        if self.reload:
            for files_changed in self:
                if files_changed:
                    logger.warning(
                        'Detected changes in %s. Reloading...',
                        ', '.join(map(_display_path, files_changed)),
                    )
                    self.restart()
        else:
            self.stop_event.wait()

        self.terminate()

    def start(self):
        for _ in range(self.workers):
            process = start_process(
                consume,
                'function',
                (),
                self.consume_kwargs,
            )
            self.pool.append(process)
            logger.info(f'Consumer process with pid {process.pid} started')

    def terminate(self, *args, **kwargs):
        while self.pool:
            process = self.pool.pop()
            process.stop(sigint_timeout=self.sigint_timeout, sigkill_timeout=self.sigkill_timeout)
            logger.info(f'Consumer process with pid {process.pid} stopped.')

    def restart(self, *args, **kwargs):
        self.terminate()
        self.start()

    def __iter__(self):
        return self

    def __next__(self):
        changes = next(self.watcher)
        if changes:
            return list({Path(c[1]) for c in changes})
        return None


class Command(BaseCommand):
    help = 'Starts CQRS worker, which consumes messages from message queue.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--workers',
            '-w',
            help='Number of workers',
            type=int,
            default=1,
        )
        parser.add_argument(
            '--cqrs-id',
            '-cid',
            nargs='*',
            type=str,
            help='Choose model(s) by CQRS_ID for consuming',
        )
        parser.add_argument(
            '--reload',
            '-r',
            help=(
                'Enable reload signal SIGHUP and autoreload '
                'on file changes'
            ),
            action='store_true',
            default=False,
        )
        parser.add_argument(
            '--ignore-paths',
            nargs='?',
            type=str,
            help=(
                'Specify directories to ignore, '
                'to ignore multiple paths use a comma as separator, '
                'e.g. "env" or "env,node_modules"'
            ),
        )
        parser.add_argument(
            '--sigint-timeout',
            nargs='?',
            type=int,
            default=5,
            help='How long to wait for the sigint timeout before sending sigkill.',
        )
        parser.add_argument(
            '--sigkill-timeout',
            nargs='?',
            type=int,
            default=1,
            help='How long to wait for the sigkill timeout before issuing a timeout exception.',
        )

    def handle(
            self,
            *args,
            workers=1,
            cqrs_id=None,
            reload=False,
            ignore_paths=None,
            sigint_timeout=5,
            sigkill_timeout=1,
            **options,
    ):

        paths_to_ignore = None
        if ignore_paths:
            paths_to_ignore = [Path(p).resolve() for p in ignore_paths.split(',')]

        workers_manager = WorkersManager(
            workers=workers,
            consume_kwargs=self.get_consume_kwargs(cqrs_id),
            reload=reload,
            ignore_paths=paths_to_ignore,
            sigint_timeout=sigint_timeout,
            sigkill_timeout=sigkill_timeout,
        )

        workers_manager.run()

    def get_consume_kwargs(self, ids_list):
        consume_kwargs = {}
        if ids_list:
            cqrs_ids = set()
            for cqrs_id in ids_list:
                model = ReplicaRegistry.get_model_by_cqrs_id(cqrs_id)
                if not model:
                    raise CommandError('Wrong CQRS ID: {0}!'.format(cqrs_id))

                cqrs_ids.add(cqrs_id)

            consume_kwargs['cqrs_ids'] = cqrs_ids

        return consume_kwargs



#  Copyright © 2023 Ingram Micro Inc. All rights reserved.

from dateutil.parser import parse as dateutil_parse
from django.utils import timezone

from dj_cqrs.correlation import get_correlation_id
from dj_cqrs.utils import get_json_valid_value, get_message_expiration_dt


class TransportPayload:
    """Transport message payload.

    Args:
        signal_type (dj_cqrs.constants.SignalType): Type of the signal for this message.
        cqrs_id (str): The unique CQRS identifier of the model.
        instance_data (dict): Serialized data of the instance that
            generates the event.
        instance_pk (str): Primary key of the instance.
        queue (str): Queue to synchronize, defaults to None.
        previous_data (dict): Previous values for fields tracked for changes,
            defaults to None.
        correlation_id (str): Correlation ID of process, where this payload is used.
        retries (int): Current number of message retries.
        expires (datetime): Message expiration datetime, infinite if None
        meta (dict): Payload metadata
    """

    def __init__(
        self,
        signal_type,
        cqrs_id: str,
        instance_data: dict,
        instance_pk: str,
        queue: str = None,
        previous_data: dict = None,
        correlation_id: str = None,
        expires=None,
        retries: int = 0,
        meta: dict = None,
    ):
        self.__signal_type = signal_type
        self.__cqrs_id = cqrs_id
        self.__instance_data = instance_data
        self.__instance_pk = instance_pk
        self.__queue = queue
        self.__previous_data = previous_data
        self.__meta = meta

        if correlation_id:
            self.__correlation_id = correlation_id
        else:
            self.__correlation_id = get_correlation_id(signal_type, cqrs_id, instance_pk, queue)

        self.__expires = expires
        self.__retries = retries

    @classmethod
    def from_message(cls, dct):
        """Builds payload from message data.

        Args:
            dct (dict): Deserialized message body data.

        Returns:
            (TransportPayload): TransportPayload instance.
        """
        if 'expires' in dct:
            expires = dct['expires']
            if dct['expires'] is not None:
                expires = dateutil_parse(dct['expires'])
        else:
            # Backward compatibility for old messages otherwise they are infinite by default.
            expires = get_message_expiration_dt()

        return cls(
            dct['signal_type'],
            dct['cqrs_id'],
            dct['instance_data'],
            dct.get('instance_pk'),
            previous_data=dct.get('previous_data'),
            correlation_id=dct.get('correlation_id'),
            expires=expires,
            retries=dct.get('retries') or 0,
            meta=dct.get('meta'),
        )

    @property
    def signal_type(self):
        return self.__signal_type

    @property
    def cqrs_id(self):
        return self.__cqrs_id

    @property
    def instance_data(self):
        return self.__instance_data

    @property
    def pk(self):
        return self.__instance_pk

    @property
    def queue(self):
        return self.__queue

    @property
    def previous_data(self):
        return self.__previous_data

    @property
    def correlation_id(self):
        return self.__correlation_id

    @property
    def meta(self):
        return self.__meta

    @property
    def expires(self):
        return self.__expires

    @property
    def retries(self):
        return self.__retries

    @retries.setter
    def retries(self, value):
        assert value >= 0, "Payload retries field should be 0 or positive integer."
        self.__retries = value

    def to_dict(self) -> dict:
        """Return the payload as a dictionary.

        Returns:
            (dict): This payload.
        """
        expires = self.__expires
        if expires:
            expires = expires.replace(microsecond=0).isoformat()

        return {
            'signal_type': self.__signal_type,
            'cqrs_id': self.__cqrs_id,
            'instance_data': self.__instance_data,
            'previous_data': self.__previous_data,
            'instance_pk': get_json_valid_value(self.__instance_pk),
            'correlation_id': get_json_valid_value(self.__correlation_id),
            'retries': self.__retries,
            'expires': expires,
            'meta': self.__meta,
        }

    def is_expired(self):
        """Checks if this payload is expired.

        Returns:
            (bool): True if payload is expired, False otherwise.
        """
        return (
            self.__expires is not None
            and self.__expires <= timezone.now()
        )



#  Copyright © 2023 Ingram Micro Inc. All rights reserved.

import logging
from inspect import getfullargspec, isfunction

from django.utils.module_loading import import_string

from dj_cqrs.constants import (
    DEFAULT_MASTER_AUTO_UPDATE_FIELDS,
    DEFAULT_MASTER_MESSAGE_TTL,
    DEFAULT_REPLICA_DELAY_QUEUE_MAX_SIZE,
    DEFAULT_REPLICA_MAX_RETRIES,
    DEFAULT_REPLICA_RETRY_DELAY,
)
from dj_cqrs.registries import MasterRegistry, ReplicaRegistry
from dj_cqrs.transport import BaseTransport


logger = logging.getLogger('django-cqrs')


def validate_settings(settings):
    is_master = bool(MasterRegistry.models)
    is_replica = bool(ReplicaRegistry.models)
    if (not is_master) and (not is_replica):  # pragma: no cover
        return

    assert hasattr(settings, 'CQRS'), 'CQRS configuration must be set in Django project settings.'

    cqrs_settings = settings.CQRS
    assert isinstance(cqrs_settings, dict), 'CQRS configuration must be dict.'

    _validate_transport(cqrs_settings)

    if is_master or ('master' in cqrs_settings):
        _validate_master(cqrs_settings)

    if is_replica or ('replica' in cqrs_settings):
        _validate_replica(cqrs_settings)


def _validate_transport(cqrs_settings):
    transport_cls_location = cqrs_settings.get('transport')
    if not transport_cls_location:
        raise AssertionError('CQRS transport is not set.')

    transport = import_string(transport_cls_location)
    if not issubclass(transport, BaseTransport):
        raise AssertionError(
            'CQRS transport must be inherited from `dj_cqrs.transport.BaseTransport`.',
        )


def _validate_master(cqrs_settings):
    default_master_settings = {
        'master': {
            'CQRS_AUTO_UPDATE_FIELDS': DEFAULT_MASTER_AUTO_UPDATE_FIELDS,
            'CQRS_MESSAGE_TTL': DEFAULT_MASTER_MESSAGE_TTL,
            'correlation_function': None,
            'meta_function': None,
        },
    }

    if 'master' not in cqrs_settings:
        cqrs_settings.update(default_master_settings)
        return

    master_settings = cqrs_settings['master']
    assert isinstance(master_settings, dict), 'CQRS master configuration must be dict.'

    _validate_master_auto_update_fields(master_settings)
    _validate_master_message_ttl(master_settings)
    _validate_master_correlation_func(master_settings)
    _validate_master_meta_func(master_settings)


def _validate_master_auto_update_fields(master_settings):
    if 'CQRS_AUTO_UPDATE_FIELDS' in master_settings:
        assert isinstance(master_settings['CQRS_AUTO_UPDATE_FIELDS'], bool), (
            'CQRS master CQRS_AUTO_UPDATE_FIELDS must be bool.'
        )
    else:
        master_settings['CQRS_AUTO_UPDATE_FIELDS'] = DEFAULT_MASTER_AUTO_UPDATE_FIELDS


def _validate_master_message_ttl(master_settings):
    if 'CQRS_MESSAGE_TTL' in master_settings:
        min_message_ttl = 1
        message_ttl = master_settings['CQRS_MESSAGE_TTL']
        if (message_ttl is not None) and (
            not isinstance(message_ttl, int) or message_ttl < min_message_ttl
        ):
            # No error is raised for backward compatibility
            # TODO: raise error in 2.0.0
            logger.warning(
                'Settings CQRS_MESSAGE_TTL=%s is invalid, using default %s.',
                message_ttl, DEFAULT_MASTER_MESSAGE_TTL,
            )
            master_settings['CQRS_MESSAGE_TTL'] = DEFAULT_MASTER_MESSAGE_TTL
    else:
        master_settings['CQRS_MESSAGE_TTL'] = DEFAULT_MASTER_MESSAGE_TTL


def _validate_master_correlation_func(master_settings):
    correlation_func = master_settings.get('correlation_function')
    if not correlation_func:
        master_settings['correlation_function'] = None
    elif not callable(correlation_func):
        raise AssertionError('CQRS master correlation_function must be callable.')


def _validate_master_meta_func(master_settings):
    meta_func = master_settings.get('meta_function')
    if not meta_func:
        master_settings['meta_function'] = None
        return

    if isinstance(meta_func, str):
        try:
            meta_func = import_string(meta_func)
        except ImportError:
            raise AssertionError('CQRS master meta_function import error.')

    if not isfunction(meta_func):
        raise AssertionError('CQRS master meta_function must be function.')

    r = getfullargspec(meta_func)
    if not r.varkw:
        raise AssertionError('CQRS master meta_function must support **kwargs.')

    master_settings['meta_function'] = meta_func


def _validate_replica(cqrs_settings):
    queue = cqrs_settings.get('queue')
    assert queue, 'CQRS queue is not set.'
    assert isinstance(queue, str), 'CQRS queue must be string.'

    default_replica_settings = {
        'replica': {
            'CQRS_MAX_RETRIES': DEFAULT_REPLICA_MAX_RETRIES,
            'CQRS_RETRY_DELAY': DEFAULT_REPLICA_RETRY_DELAY,
            'delay_queue_max_size': DEFAULT_REPLICA_DELAY_QUEUE_MAX_SIZE,
        },
    }

    if 'replica' not in cqrs_settings:
        cqrs_settings.update(default_replica_settings)
        return

    replica_settings = cqrs_settings['replica']
    assert isinstance(replica_settings, dict), 'CQRS replica configuration must be dict.'

    _validate_replica_max_retries(replica_settings)
    _validate_replica_retry_delay(replica_settings)
    _validate_replica_delay_queue_max_size(replica_settings)


def _validate_replica_max_retries(replica_settings):
    if 'CQRS_MAX_RETRIES' in replica_settings:
        min_retries = 0
        max_retries = replica_settings['CQRS_MAX_RETRIES']
        if (max_retries is not None) and (
            not isinstance(max_retries, int) or max_retries < min_retries
        ):
            # No error is raised for backward compatibility
            # TODO: raise error in 2.0.0
            logger.warning(
                'Replica setting CQRS_MAX_RETRIES=%s is invalid, using default %s.',
                max_retries, DEFAULT_REPLICA_MAX_RETRIES,
            )
            replica_settings['CQRS_MAX_RETRIES'] = DEFAULT_REPLICA_MAX_RETRIES
    else:
        replica_settings['CQRS_MAX_RETRIES'] = DEFAULT_REPLICA_MAX_RETRIES


def _validate_replica_retry_delay(replica_settings):
    min_retry_delay = 0
    retry_delay = replica_settings.get('CQRS_RETRY_DELAY')
    if 'CQRS_RETRY_DELAY' not in replica_settings:
        replica_settings['CQRS_RETRY_DELAY'] = DEFAULT_REPLICA_RETRY_DELAY
    elif not isinstance(retry_delay, int) or retry_delay < min_retry_delay:
        # No error is raised for backward compatibility
        # TODO: raise error in 2.0.0
        logger.warning(
            'Replica setting CQRS_RETRY_DELAY=%s is invalid, using default %s.',
            retry_delay, DEFAULT_REPLICA_RETRY_DELAY,
        )
        replica_settings['CQRS_RETRY_DELAY'] = DEFAULT_REPLICA_RETRY_DELAY


def _validate_replica_delay_queue_max_size(replica_settings):
    min_qsize = 0
    max_qsize = replica_settings.get('delay_queue_max_size')
    if 'delay_queue_max_size' not in replica_settings:
        max_qsize = DEFAULT_REPLICA_DELAY_QUEUE_MAX_SIZE
    elif (max_qsize is not None) and (not isinstance(max_qsize, int) or max_qsize <= min_qsize):
        # No error is raised for backward compatibility
        # TODO: raise error in 2.0.0
        logger.warning(
            'Settings delay_queue_max_size=%s is invalid, using default %s.',
            max_qsize, DEFAULT_REPLICA_DELAY_QUEUE_MAX_SIZE,
        )
        max_qsize = DEFAULT_REPLICA_DELAY_QUEUE_MAX_SIZE

    replica_settings['delay_queue_max_size'] = max_qsize



#  Copyright © 2023 Ingram Micro Inc. All rights reserved.

import ujson
from django.core.exceptions import FieldError
from django.core.management.base import BaseCommand, CommandError
from django.utils.timezone import now

from dj_cqrs.management.utils import batch_qs
from dj_cqrs.registries import ReplicaRegistry


class Command(BaseCommand):
    help = 'Streaming diff of CQRS model pks from replica service to check for deleted objects.'

    @classmethod
    def serialize_package(cls, package):
        return ujson.dumps(package)

    def add_arguments(self, parser):
        parser.add_argument(
            '--cqrs-id', '-cid',
            help='CQRS_ID of the replica model',
            type=str,
            required=True,
        )
        parser.add_argument(
            '--filter', '-f',
            help='Filter kwargs',
            type=str,
            default=None,
        )
        parser.add_argument(
            '--batch', '-b',
            help='Batch size',
            type=int,
            default=10000,
        )

    def handle(self, *args, **options):
        model = self._get_model(options)
        batch_size = self._get_batch_size(options)

        qs = model._default_manager.values().order_by()
        if options['filter']:
            try:
                kwargs = ujson.loads(options['filter'])
                if not isinstance(kwargs, dict):
                    raise ValueError
            except ValueError:
                raise CommandError('Bad filter kwargs!')

            try:
                qs = qs.filter(**kwargs)
            except FieldError as e:
                raise CommandError('Bad filter kwargs! {0}'.format(str(e)))

        if not qs.exists():
            self.stderr.write('No objects found for filter!')
            return

        current_dt = now()
        self.stdout.write('{0},{1}'.format(model.CQRS_ID, str(current_dt)))

        for bqs in batch_qs(qs.values_list('pk', flat=True), batch_size=batch_size):
            self.stdout.write(self.serialize_package(list(bqs)))

    @staticmethod
    def _get_model(options):
        cqrs_id = options['cqrs_id']
        model = ReplicaRegistry.get_model_by_cqrs_id(cqrs_id)

        if not model:
            raise CommandError('Wrong CQRS ID: {0}!'.format(cqrs_id))

        return model

    @staticmethod
    def _get_batch_size(options):
        return options['batch']



#  Copyright © 2023 Ingram Micro Inc. All rights reserved.

import logging
import time
from datetime import timedelta
from socket import gaierror
from urllib.parse import unquote, urlparse

import ujson
from django.conf import settings
from django.utils import timezone
from pika import (
    BasicProperties,
    BlockingConnection,
    ConnectionParameters,
    credentials,
    exceptions,
)
from pika.adapters.utils.connection_workflow import AMQPConnectorException

from dj_cqrs.constants import DEFAULT_DEAD_MESSAGE_TTL, SignalType
from dj_cqrs.controller import consumer
from dj_cqrs.dataclasses import TransportPayload
from dj_cqrs.delay import DelayMessage, DelayQueue
from dj_cqrs.registries import ReplicaRegistry
from dj_cqrs.transport import BaseTransport
from dj_cqrs.transport.mixins import LoggingMixin
from dj_cqrs.utils import get_delay_queue_max_size, get_messages_prefetch_count_per_worker


logger = logging.getLogger('django-cqrs')


class RabbitMQTransport(LoggingMixin, BaseTransport):
    """Transport class for RabbitMQ."""
    CONSUMER_RETRY_TIMEOUT = 5
    PRODUCER_RETRIES = 1

    _producer_connection = None
    _producer_channel = None

    @classmethod
    def clean_connection(cls):
        """Clean the RabbitMQ connection."""
        connection = cls._producer_connection
        if connection and not connection.is_closed:
            try:
                connection.close()
            except (exceptions.StreamLostError, exceptions.ConnectionClosed, ConnectionError):
                logger.warning("Connection was closed or is closing. Skip it...")

        cls._producer_connection = None
        cls._producer_channel = None

    @classmethod
    def consume(cls, cqrs_ids=None):
        """Receive data from master model.

        Args:
            cqrs_ids (str): cqrs ids.
        """
        consumer_rabbit_settings = cls._get_consumer_settings()
        common_rabbit_settings = cls._get_common_settings()

        while True:
            connection = None
            try:
                delay_queue = DelayQueue(max_size=get_delay_queue_max_size())
                connection, channel, consumer_generator = cls._get_consumer_rmq_objects(
                    *(common_rabbit_settings + consumer_rabbit_settings), cqrs_ids=cqrs_ids,
                )

                for method_frame, properties, body in consumer_generator:
                    if method_frame is not None:
                        cls._consume_message(
                            channel, method_frame, properties, body, delay_queue,
                        )
                    cls._process_delay_messages(channel, delay_queue)
            except (exceptions.AMQPError,
                    exceptions.ChannelError,
                    exceptions.ReentrancyError,
                    gaierror):
                logger.warning('AMQP connection error. Reconnecting...', exc_info=True)
                time.sleep(cls.CONSUMER_RETRY_TIMEOUT)
            finally:
                if connection and not connection.is_closed:
                    connection.close()

    @classmethod
    def produce(cls, payload):
        """
        Send data from master model to replicas.

        Args:
            payload (dj_cqrs.dataclasses.TransportPayload): Transport payload from master model.
        """
        cls._produce_with_retries(payload, retries=cls.PRODUCER_RETRIES)

    @classmethod
    def _produce_with_retries(cls, payload, retries):
        try:
            rmq_settings = cls._get_common_settings()
            exchange = rmq_settings[-1]
            # Decided not to create context-manager to stay within the class
            _, channel = cls._get_producer_rmq_objects(
                *rmq_settings, signal_type=payload.signal_type,
            )

            cls._produce_message(channel, exchange, payload)
            cls.log_produced(payload)
        except (
            exceptions.AMQPError,
            exceptions.ChannelError,
            exceptions.ReentrancyError,
            AMQPConnectorException,
            AssertionError,
        ) as e:
            # in case of any error - close connection and try to reconnect
            cls.clean_connection()

            base_log_message = "CQRS couldn't be published: pk = {0} ({1}).".format(
                payload.pk, payload.cqrs_id,
            )
            if not retries:
                logger.exception(base_log_message)
                return

            logger.warning('{0} Error: {1}. Reconnect...'.format(
                base_log_message, e.__class__.__name__,
            ))

            cls._produce_with_retries(payload, retries - 1)

    @classmethod
    def _consume_message(cls, ch, method, properties, body, delay_queue):
        try:
            dct = ujson.loads(body)
        except ValueError:
            logger.error("CQRS couldn't be parsed: {0}.".format(body))
            ch.basic_reject(delivery_tag=method.delivery_tag, requeue=False)
            return

        required_keys = {'instance_pk', 'signal_type', 'cqrs_id', 'instance_data'}
        for key in required_keys:
            if key not in dct:
                msg = "CQRS couldn't proceed, %s isn't found in body: %s."
                logger.error(msg, key, body)
                ch.basic_reject(delivery_tag=method.delivery_tag, requeue=False)
                return

        payload = TransportPayload.from_message(dct)
        cls.log_consumed(payload)

        delivery_tag = method.delivery_tag
        if payload.is_expired():
            cls._add_to_dead_letter_queue(ch, payload)
            cls._nack(ch, delivery_tag)
            return

        instance, exception = None, None
        try:
            instance = consumer.consume(payload)
        except Exception as e:
            exception = e
            logger.error("CQRS service exception", exc_info=True)

        if instance and exception is None:
            cls._ack(ch, delivery_tag, payload)
        else:
            cls._fail_message(
                ch, delivery_tag, payload, exception, delay_queue,
            )

    @classmethod
    def _fail_message(cls, channel, delivery_tag, payload, exception, delay_queue):
        cls.log_consumed_failed(payload)
        model_cls = ReplicaRegistry.get_model_by_cqrs_id(payload.cqrs_id)
        if model_cls is None:
            logger.error("Model for cqrs_id {0} is not found.".format(payload.cqrs_id))
            cls._nack(channel, delivery_tag)
            return

        if model_cls.should_retry_cqrs(payload.retries, exception):
            delay = model_cls.get_cqrs_retry_delay(payload.retries)
            cls._delay_message(channel, delivery_tag, payload, delay, delay_queue)
        else:
            cls._add_to_dead_letter_queue(channel, payload)
            cls._nack(channel, delivery_tag)

    @classmethod
    def _delay_message(cls, channel, delivery_tag, payload, delay, delay_queue):
        if delay_queue.full():
            # Memory limits handling, requeuing message with lowest ETA
            requeue_message = delay_queue.get()
            cls._requeue_message(
                channel,
                requeue_message.delivery_tag,
                requeue_message.payload,
            )

        eta = timezone.now() + timedelta(seconds=delay)
        delay_message = DelayMessage(delivery_tag, payload, eta)
        delay_queue.put(delay_message)
        cls.log_delayed(payload, delay, delay_message.eta)

    @classmethod
    def _add_to_dead_letter_queue(cls, channel, payload):
        replica_settings = settings.CQRS.get('replica', {})
        dead_message_ttl = DEFAULT_DEAD_MESSAGE_TTL
        if 'dead_message_ttl' in replica_settings:
            dead_message_ttl = replica_settings['dead_message_ttl']

        expiration = None
        if dead_message_ttl is not None:
            expiration = str(dead_message_ttl * 1000)  # milliseconds

        payload.is_dead_letter = True
        exchange = cls._get_common_settings()[-1]
        cls._produce_message(channel, exchange, payload, expiration)
        cls.log_dead_letter(payload)

    @classmethod
    def _requeue_message(cls, channel, delivery_tag, payload):
        payload.retries += 1
        payload.is_requeue = True

        cls.produce(payload)
        cls._nack(channel, delivery_tag)
        cls.log_requeued(payload)

    @classmethod
    def _process_delay_messages(cls, channel, delay_queue):
        for delay_message in delay_queue.get_ready():
            cls._requeue_message(channel, delay_message.delivery_tag, delay_message.payload)

    @classmethod
    def _produce_message(cls, channel, exchange, payload, expiration=None):
        routing_key = cls._get_produced_message_routing_key(payload)

        channel.basic_publish(
            exchange=exchange,
            routing_key=routing_key,
            body=ujson.dumps(payload.to_dict()),
            mandatory=True,
            properties=BasicProperties(
                content_type='text/plain',
                delivery_mode=2,  # make message persistent
                expiration=expiration,
            ),
        )

    @classmethod
    def _get_produced_message_routing_key(cls, payload):
        routing_key = payload.cqrs_id

        if payload.signal_type == SignalType.SYNC and payload.queue:
            routing_key = 'cqrs.{0}.{1}'.format(payload.queue, routing_key)
        elif getattr(payload, 'is_dead_letter', False):
            dead_letter_queue_name = cls._get_consumer_settings()[1]
            routing_key = 'cqrs.{0}.{1}'.format(dead_letter_queue_name, routing_key)
        elif getattr(payload, 'is_requeue', False):
            queue = cls._get_consumer_settings()[0]
            routing_key = 'cqrs.{0}.{1}'.format(queue, routing_key)

        return routing_key

    @classmethod
    def _get_consumer_rmq_objects(
        cls,
        host,
        port,
        creds,
        exchange,
        queue_name,
        dead_letter_queue_name,
        prefetch_count,
        cqrs_ids=None,
    ):
        connection = BlockingConnection(
            ConnectionParameters(host=host, port=port, credentials=creds),
        )
        channel = connection.channel()
        channel.basic_qos(prefetch_count=prefetch_count)
        cls._declare_exchange(channel, exchange)

        channel.queue_declare(queue_name, durable=True, exclusive=False)
        channel.queue_declare(dead_letter_queue_name, durable=True, exclusive=False)

        for cqrs_id, _ in ReplicaRegistry.models.items():
            if cqrs_ids and cqrs_id not in cqrs_ids:
                continue

            channel.queue_bind(exchange=exchange, queue=queue_name, routing_key=cqrs_id)

            # Every service must have specific SYNC or requeue routes
            channel.queue_bind(
                exchange=exchange,
                queue=queue_name,
                routing_key='cqrs.{0}.{1}'.format(queue_name, cqrs_id),
            )

            # Dead letter
            channel.queue_bind(
                exchange=exchange,
                queue=dead_letter_queue_name,
                routing_key='cqrs.{0}.{1}'.format(dead_letter_queue_name, cqrs_id),
            )

        delay_queue_check_timeout = 1  # seconds
        consumer_generator = channel.consume(
            queue=queue_name,
            auto_ack=False,
            exclusive=False,
            inactivity_timeout=delay_queue_check_timeout,
        )
        return connection, channel, consumer_generator

    @classmethod
    def _get_producer_rmq_objects(cls, host, port, creds, exchange, signal_type=None):
        """
        Use shared connection in case of sync mode, otherwise create new connection for each
        message
        """
        if signal_type == SignalType.SYNC:
            if cls._producer_connection is None:
                connection, channel = cls._create_connection(host, port, creds, exchange)

                cls._producer_connection = connection
                cls._producer_channel = channel

            return cls._producer_connection, cls._producer_channel
        else:
            return cls._create_connection(host, port, creds, exchange)

    @classmethod
    def _create_connection(cls, host, port, creds, exchange):
        connection = BlockingConnection(
            ConnectionParameters(
                host=host,
                port=port,
                credentials=creds,
                blocked_connection_timeout=10,
            ),
        )
        channel = connection.channel()
        channel.basic_qos(prefetch_count=get_messages_prefetch_count_per_worker())
        cls._declare_exchange(channel, exchange)

        return connection, channel

    @staticmethod
    def _declare_exchange(channel, exchange):
        channel.exchange_declare(
            exchange=exchange,
            exchange_type='topic',
            durable=True,
        )

    @staticmethod
    def _parse_url(url):
        scheme = urlparse(url).scheme
        assert scheme == 'amqp', 'Scheme must be "amqp" for RabbitMQTransport.'

        schemeless = url[len(scheme) + 3:]
        parts = urlparse('http://' + schemeless)

        return (
            unquote(parts.hostname or '') or ConnectionParameters.DEFAULT_HOST,
            parts.port or ConnectionParameters.DEFAULT_PORT,
            unquote(parts.username or '') or ConnectionParameters.DEFAULT_USERNAME,
            unquote(parts.password or '') or ConnectionParameters.DEFAULT_PASSWORD,
        )

    @classmethod
    def _get_common_settings(cls):
        if 'url' in settings.CQRS:
            host, port, user, password = cls._parse_url(settings.CQRS.get('url'))
        else:
            host = settings.CQRS.get('host', ConnectionParameters.DEFAULT_HOST)
            port = settings.CQRS.get('port', ConnectionParameters.DEFAULT_PORT)
            user = settings.CQRS.get('user', ConnectionParameters.DEFAULT_USERNAME)
            password = settings.CQRS.get('password', ConnectionParameters.DEFAULT_PASSWORD)
        exchange = settings.CQRS.get('exchange', 'cqrs')
        return (
            host,
            port,
            credentials.PlainCredentials(user, password, erase_on_connect=True),
            exchange,
        )

    @staticmethod
    def _get_consumer_settings():
        queue_name = settings.CQRS['queue']

        replica_settings = settings.CQRS.get('replica', {})
        dead_letter_queue_name = 'dead_letter_{0}'.format(queue_name)
        if 'dead_letter_queue' in replica_settings:
            dead_letter_queue_name = replica_settings['dead_letter_queue']

        if 'consumer_prefetch_count' in settings.CQRS:
            logger.warning(
                "The 'consumer_prefetch_count' setting is ignored for RabbitMQTransport.",
            )
        prefetch_count = get_messages_prefetch_count_per_worker()

        return (
            queue_name,
            dead_letter_queue_name,
            prefetch_count,
        )

    @classmethod
    def _ack(cls, channel, delivery_tag, payload=None):
        channel.basic_ack(delivery_tag)
        if payload is not None:
            cls.log_consumed_accepted(payload)

    @classmethod
    def _nack(cls, channel, delivery_tag, payload=None):
        channel.basic_nack(delivery_tag, requeue=False)
        if payload is not None:
            cls.log_consumed_denied(payload)



#  Copyright © 2023 Ingram Micro Inc. All rights reserved.

import logging

from django.core.exceptions import ValidationError
from django.db import Error, transaction
from django.db.models import F, Manager
from django.utils import timezone

from dj_cqrs.constants import FIELDS_TRACKER_FIELD_NAME, TRACKED_FIELDS_ATTR_NAME


logger = logging.getLogger('django-cqrs')


class MasterManager(Manager):
    def bulk_create(self, objs, **kwargs):
        """
        Custom bulk create method to support sending of create signals.
        This can be used only in cases, when IDs are generated on client or DB returns IDs.

        Args:
            objs (List[django.db.models.Model]): List of objects for creation.
            kwargs (dict): Bulk create kwargs.
        """
        for obj in objs:
            obj.save_tracked_fields()
        objs = super().bulk_create(objs, **kwargs)

        if objs:
            self.model.call_post_bulk_create(objs, using=self.db)

        return objs

    def bulk_update(self, queryset, **kwargs):
        """ Custom update method to support sending of update signals.

        Args:
            queryset (django.db.models.QuerySet): Django Queryset (f.e. filter).
            kwargs (dict): Update kwargs.
        """
        prev_data_mapper = {}
        collect_prev_data = hasattr(self.model, FIELDS_TRACKER_FIELD_NAME)

        # Add filter by list of ids in case of update kwargs
        # are the same as the chain filter kwargs in the Queryset.
        # If that happen the .all() method will refresh after update and
        # result in an empty Queryset that will not send the signal.
        ids_list = list(queryset.values_list('pk', flat=True))

        def list_all():
            return list(queryset.model.objects.filter(pk__in=ids_list).all())

        with transaction.atomic(savepoint=False):
            if collect_prev_data:
                objs = list_all()
                if not objs:
                    return

                for obj in objs:
                    prev_data_mapper[obj.pk] = getattr(obj, FIELDS_TRACKER_FIELD_NAME).current()

            current_dt = timezone.now()
            result = queryset.update(
                cqrs_revision=F('cqrs_revision') + 1, cqrs_updated=current_dt, **kwargs,
            )

            objs = list_all()
            if collect_prev_data:
                for obj in objs:
                    setattr(obj, TRACKED_FIELDS_ATTR_NAME, prev_data_mapper.get(obj.pk))

        queryset.model.call_post_update(objs, using=queryset.db)

        return result


class ReplicaManager(Manager):
    def save_instance(
        self,
        master_data: dict,
        previous_data: dict = None,
        sync: bool = False,
        meta: dict = None,
    ):
        """ This method saves (creates or updates) model instance from CQRS master instance data.

        Args:
            master_data (dict): CQRS master instance data.
            previous_data (dict): Previous values for tracked fields.
            sync (bool): Sync package flag.
            meta (dict): Payload metadata, if exists.

        Returns:
            (django.db.models.Model): Model instance.
        """
        mapped_data = self._map_save_data(master_data)
        mapped_previous_data = self._map_previous_data(previous_data) if previous_data else None
        if mapped_data:
            pk_name = self._get_model_pk_name()
            pk_value = mapped_data[pk_name]
            f_kwargs = {pk_name: pk_value}

            qs = self.model._default_manager.filter(**f_kwargs).order_by()
            if self.model.CQRS_SELECT_FOR_UPDATE:
                qs = qs.select_for_update()

            instance = qs.first()

            if instance:
                return self.update_instance(
                    instance,
                    mapped_data,
                    previous_data=mapped_previous_data,
                    sync=sync,
                    meta=meta,
                )

            return self.create_instance(
                mapped_data,
                previous_data=mapped_previous_data,
                sync=sync,
                meta=meta,
            )

    def create_instance(
        self,
        mapped_data: dict,
        previous_data: dict = None,
        sync: bool = False,
        meta: dict = None,
    ):
        """ This method creates model instance from mapped CQRS master instance data.

        Args:
            mapped_data (dict): Mapped CQRS master instance data.
            previous_data (dict): Previous values for tracked fields.
            sync (bool): Sync package flag.
            meta (dict): Payload metadata, if exists.

        Returns:
            (django.db.models.Model): ReplicaMixin instance.
        """
        f_kw = {'previous_data': previous_data}
        if self.model.CQRS_META:
            f_kw['meta'] = meta

        try:
            return self.model.cqrs_create(sync, mapped_data, **f_kw)
        except (Error, ValidationError) as e:
            pk_value = mapped_data[self._get_model_pk_name()]

            logger.error(
                '{0}\nCQRS create error: pk = {1} ({2}).'.format(
                    str(e), pk_value, self.model.CQRS_ID,
                ),
            )

    def update_instance(
        self,
        instance,
        mapped_data: dict,
        previous_data: dict = None,
        sync: bool = False,
        meta: dict = None,
    ):
        """ This method updates model instance from mapped CQRS master instance data.

        Args:
            instance (django.db.models.Model): ReplicaMixin model instance.
            mapped_data (dict): Mapped CQRS master instance data.
            previous_data (dict): Previous values for tracked fields.
            sync (bool): Sync package flag.
            meta (dict): Payload metadata, if exists.

        Returns:
            (django.db.models.Model): ReplicaMixin instance.
        """
        pk_value = mapped_data[self._get_model_pk_name()]
        current_cqrs_revision = mapped_data['cqrs_revision']
        existing_cqrs_revision = instance.cqrs_revision

        if sync:
            if existing_cqrs_revision > current_cqrs_revision:
                w_tpl = (
                    'CQRS revision downgrade on sync: pk = {0}, '
                    'cqrs_revision = new {1} / existing {2} ({3}).'
                )
                logger.warning(w_tpl.format(
                    pk_value, current_cqrs_revision, existing_cqrs_revision, self.model.CQRS_ID,
                ))

        else:
            if existing_cqrs_revision > current_cqrs_revision:
                e_tpl = (
                    'Wrong CQRS sync order: pk = {0}, '
                    'cqrs_revision = new {1} / existing {2} ({3}).'
                )
                logger.error(e_tpl.format(
                    pk_value, current_cqrs_revision, existing_cqrs_revision, self.model.CQRS_ID,
                ))
                return instance

            if existing_cqrs_revision == current_cqrs_revision:
                logger.error(
                    'Received duplicate CQRS data: pk = {0}, cqrs_revision = {1} ({2}).'.format(
                        pk_value, current_cqrs_revision, self.model.CQRS_ID,
                    ),
                )
                if current_cqrs_revision == 0:
                    logger.warning(
                        'CQRS potential creation race condition: pk = {0} ({1}).'.format(
                            pk_value, self.model.CQRS_ID,
                        ),
                    )

                return instance

            if current_cqrs_revision != instance.cqrs_revision + 1:
                w_tpl = (
                    'Lost or filtered out {0} CQRS packages: pk = {1}, cqrs_revision = {2} ({3})'
                )
                logger.warning(w_tpl.format(
                    current_cqrs_revision - instance.cqrs_revision - 1,
                    pk_value, current_cqrs_revision, self.model.CQRS_ID,
                ))

        f_kw = {'previous_data': previous_data}
        if self.model.CQRS_META:
            f_kw['meta'] = meta

        try:
            return instance.cqrs_update(sync, mapped_data, **f_kw)
        except (Error, ValidationError) as e:
            logger.error(
                '{0}\nCQRS update error: pk = {1}, cqrs_revision = {2} ({3}).'.format(
                    str(e), pk_value, current_cqrs_revision, self.model.CQRS_ID,
                ),
            )

    def delete_instance(self, master_data: dict) -> bool:
        """ This method deletes model instance from mapped CQRS master instance data.

        Args:
            master_data (dict): CQRS master instance data.

        Returns:
            Flag, if delete operation is successful (even if nothing was deleted).
        """
        mapped_data = self._map_delete_data(master_data)

        if mapped_data:
            pk_name = self._get_model_pk_name()
            pk_value = mapped_data[pk_name]
            try:
                self.model._default_manager.filter(**{pk_name: pk_value}).delete()
                return True
            except Error as e:
                logger.error(
                    '{0}\nCQRS delete error: pk = {1} ({2}).'.format(
                        str(e), pk_value, self.model.CQRS_ID,
                    ),
                )

        return False

    def _map_previous_data(self, previous_data):
        if self.model.CQRS_MAPPING is None:
            return previous_data

        mapped_previous_data = {}

        for master_name, replica_name in self.model.CQRS_MAPPING.items():
            if master_name not in previous_data:
                continue

            mapped_previous_data[replica_name] = previous_data[master_name]
        mapped_previous_data = self._remove_excessive_data(mapped_previous_data)
        return mapped_previous_data

    def _map_save_data(self, master_data):
        if not self._cqrs_fields_are_filled(master_data):
            return

        mapped_data = self._make_initial_mapping(master_data)
        if not mapped_data:
            return

        if self._get_model_pk_name() not in mapped_data:
            self._log_pk_data_error()
            return

        if self.model.CQRS_CUSTOM_SERIALIZATION:
            return mapped_data

        mapped_data = self._remove_excessive_data(mapped_data)

        if self._all_required_fields_are_filled(mapped_data):
            return mapped_data

    def _make_initial_mapping(self, master_data):
        if self.model.CQRS_MAPPING is None:
            return master_data

        mapped_data = {
            'cqrs_revision': master_data['cqrs_revision'],
            'cqrs_updated': master_data['cqrs_updated'],
        }
        for master_name, replica_name in self.model.CQRS_MAPPING.items():
            if master_name not in master_data:
                logger.error('Bad master-replica mapping for {0} ({1}).'.format(
                    master_name, self.model.CQRS_ID,
                ))
                return

            mapped_data[replica_name] = master_data[master_name]
        return mapped_data

    def _remove_excessive_data(self, data):
        opts = self.model._meta
        possible_field_names = {
            f.name for f in opts.fields
        }
        return {k: v for k, v in data.items() if k in possible_field_names}

    def _all_required_fields_are_filled(self, mapped_data):
        opts = self.model._meta

        required_field_names = {
            f.name for f in opts.fields if not f.null
        }
        if not (required_field_names - set(mapped_data.keys())):
            return True

        logger.error(
            'Not all required CQRS fields are provided in data ({0}).'.format(self.model.CQRS_ID),
        )
        return False

    def _map_delete_data(self, master_data):
        if 'id' not in master_data:
            self._log_pk_data_error()
            return

        if not self._cqrs_fields_are_filled(master_data):
            return

        return {
            self._get_model_pk_name(): master_data['id'],
            'cqrs_revision': master_data['cqrs_revision'],
            'cqrs_updated': master_data['cqrs_updated'],
        }

    def _cqrs_fields_are_filled(self, data):
        if 'cqrs_revision' in data and 'cqrs_updated' in data:
            return True

        logger.error('CQRS sync fields are not provided in data ({0}).'.format(self.model.CQRS_ID))
        return False

    def _log_pk_data_error(self):
        logger.error('CQRS PK is not provided in data ({0}).'.format(self.model.CQRS_ID))

    def _get_model_pk_name(self):
        return self.model._meta.pk.name



#  Copyright © 2023 Ingram Micro Inc. All rights reserved.

import logging

import ujson
from django.conf import settings
from kombu import (
    Connection,
    Exchange,
    Producer,
    Queue,
)
from kombu.exceptions import KombuError
from kombu.mixins import ConsumerMixin

from dj_cqrs.constants import SignalType
from dj_cqrs.controller import consumer
from dj_cqrs.dataclasses import TransportPayload
from dj_cqrs.registries import ReplicaRegistry
from dj_cqrs.transport import BaseTransport
from dj_cqrs.transport.mixins import LoggingMixin


logger = logging.getLogger('django-cqrs')


class _KombuConsumer(ConsumerMixin):

    def __init__(self, url, exchange_name, queue_name, prefetch_count, callback, cqrs_ids=None):
        self.connection = Connection(url)
        self.exchange = Exchange(
            exchange_name,
            type='topic',
            durable=True,
        )
        self.queue_name = queue_name
        self.prefetch_count = prefetch_count
        self.callback = callback
        self.queues = []
        self.cqrs_ids = cqrs_ids

        self._init_queues()

    def _init_queues(self):
        channel = self.connection.channel()
        for cqrs_id in ReplicaRegistry.models.keys():
            if (not self.cqrs_ids) or (cqrs_id in self.cqrs_ids):
                q = Queue(
                    self.queue_name,
                    exchange=self.exchange,
                    routing_key=cqrs_id,
                )
                q.maybe_bind(channel)
                q.declare()
                self.queues.append(q)

                sync_q = Queue(
                    self.queue_name,
                    exchange=self.exchange,
                    routing_key='cqrs.{0}.{1}'.format(self.queue_name, cqrs_id),
                )
                sync_q.maybe_bind(channel)
                sync_q.declare()
                self.queues.append(sync_q)

    def get_consumers(self, Consumer, channel):
        return [
            Consumer(
                queues=self.queues,
                callbacks=[self.callback],
                prefetch_count=self.prefetch_count,
                auto_declare=True,
            ),
        ]


class KombuTransport(LoggingMixin, BaseTransport):
    """Transport class for Kombu."""
    CONSUMER_RETRY_TIMEOUT = 5

    @classmethod
    def clean_connection(cls):
        """Nothing to do here"""
        pass

    @classmethod
    def consume(cls, cqrs_ids=None):
        """Receive data from master model.

        Args:
            cqrs_ids (str): cqrs ids.
        """
        queue_name, prefetch_count = cls._get_consumer_settings()
        url, exchange_name = cls._get_common_settings()

        consumer = _KombuConsumer(
            url,
            exchange_name,
            queue_name,
            prefetch_count,
            cls._consume_message,
            cqrs_ids=cqrs_ids,
        )
        consumer.run()

    @classmethod
    def produce(cls, payload):
        """
        Send data from master model to replicas.

        Args:
            payload (dj_cqrs.dataclasses.TransportPayload): Transport payload from master model.
        """
        url, exchange_name = cls._get_common_settings()

        connection = None
        try:
            # Decided not to create context-manager to stay within the class
            connection, channel = cls._get_producer_kombu_objects(url, exchange_name)
            exchange = cls._create_exchange(exchange_name)
            cls._produce_message(channel, exchange, payload)
            cls.log_produced(payload)
        except KombuError:
            logger.error("CQRS couldn't be published: pk = {0} ({1}).".format(
                payload.pk, payload.cqrs_id,
            ))
        finally:
            if connection:
                connection.close()

    @classmethod
    def _consume_message(cls, body, message):
        try:
            dct = ujson.loads(body)
        except ValueError:
            logger.error("CQRS couldn't be parsed: {0}.".format(body))
            message.reject()
            return

        required_keys = {'instance_pk', 'signal_type', 'cqrs_id', 'instance_data'}
        for key in required_keys:
            if key not in dct:
                msg = "CQRS couldn't proceed, %s isn't found in body: %s."
                logger.error(msg, key, body)
                message.reject()
                return

        payload = TransportPayload(
            dct['signal_type'],
            dct['cqrs_id'],
            dct['instance_data'],
            dct.get('instance_pk'),
            previous_data=dct.get('previous_data'),
            correlation_id=dct.get('correlation_id'),
        )

        cls.log_consumed(payload)
        instance = consumer.consume(payload)

        if instance:
            message.ack()
            cls.log_consumed_accepted(payload)
        else:
            message.reject()
            cls.log_consumed_denied(payload)

    @classmethod
    def _produce_message(cls, channel, exchange, payload):
        routing_key = cls._get_produced_message_routing_key(payload)
        producer = Producer(
            channel,
            exchange=exchange,
            auto_declare=True,
        )
        producer.publish(
            ujson.dumps(payload.to_dict()),
            routing_key=routing_key,
            mandatory=True,
            content_type='text/plain',
            delivery_mode=2,
        )

    @staticmethod
    def _get_produced_message_routing_key(payload):
        routing_key = payload.cqrs_id

        if payload.signal_type == SignalType.SYNC and payload.queue:
            routing_key = 'cqrs.{0}.{1}'.format(payload.queue, routing_key)

        return routing_key

    @classmethod
    def _get_producer_kombu_objects(cls, url, exchange_name):
        connection = Connection(url)
        channel = connection.channel()
        return connection, channel

    @staticmethod
    def _create_exchange(exchange_name):
        return Exchange(
            exchange_name,
            type='topic',
            durable=True,
        )

    @staticmethod
    def _get_common_settings():
        url = settings.CQRS.get('url', 'amqp://localhost')
        exchange = settings.CQRS.get('exchange', 'cqrs')
        return (
            url,
            exchange,
        )

    @staticmethod
    def _get_consumer_settings():
        queue_name = settings.CQRS['queue']
        consumer_prefetch_count = settings.CQRS.get('consumer_prefetch_count', 10)
        return (
            queue_name,
            consumer_prefetch_count,
        )



#  Copyright © 2023 Ingram Micro Inc. All rights reserved.

from django.conf import settings


def get_correlation_id(signal_type, cqrs_id, instance_pk, queue):
    """
    :param signal_type: Type of the signal for this message.
    :type signal_type: dj_cqrs.constants.SignalType
    :param cqrs_id: The unique CQRS identifier of the model.
    :type cqrs_id: str
    :param instance_pk: Primary key of the instance.
    :param queue: Queue to synchronize, defaults to None
    :type queue: str, optional
    """
    correlation_func = settings.CQRS.get('master', {}).get('correlation_function')
    if correlation_func:
        return correlation_func(signal_type, cqrs_id, instance_pk, queue)



#  Copyright © 2023 Ingram Micro Inc. All rights reserved.

import copy
import logging
from contextlib import ExitStack

from django.conf import settings
from django.db import Error, close_old_connections, transaction

from dj_cqrs.constants import SignalType
from dj_cqrs.registries import ReplicaRegistry


logger = logging.getLogger('django-cqrs')


def consume(payload):
    """ Consumer controller.

    :param dj_cqrs.dataclasses.TransportPayload payload: Consumed payload from master service.
    """
    payload = copy.deepcopy(payload)
    return route_signal_to_replica_model(
        payload.signal_type,
        payload.cqrs_id,
        payload.instance_data,
        previous_data=payload.previous_data,
        meta=payload.meta,
        queue=payload.queue,
    )


def route_signal_to_replica_model(
    signal_type, cqrs_id, instance_data, previous_data=None, meta=None, queue=None,
):
    """ Routes signal to model method to create/update/delete replica instance.

    :param dj_cqrs.constants.SignalType signal_type: Consumed signal type.
    :param str cqrs_id: Replica model CQRS unique identifier.
    :param dict instance_data: Master model data.
    :param dict or None previous_data: Previous model data for changed tracked fields, if exists.
    :param dict or None meta: Payload metadata, if exists.
    :param str or None queue: Synced queue.
    """
    if signal_type not in (SignalType.DELETE, SignalType.SAVE, SignalType.SYNC):
        logger.error('Bad signal type "{0}" for CQRS_ID "{1}".'.format(signal_type, cqrs_id))
        return

    model_cls = ReplicaRegistry.get_model_by_cqrs_id(cqrs_id)
    if not model_cls:
        return

    this_queue = settings.CQRS['queue']
    if signal_type == SignalType.SYNC and model_cls.CQRS_ONLY_DIRECT_SYNCS and queue != this_queue:
        return True

    db_is_needed = not model_cls.CQRS_NO_DB_OPERATIONS
    if db_is_needed:
        close_old_connections()

    is_meta_supported = model_cls.CQRS_META
    try:
        with transaction.atomic(savepoint=False) if db_is_needed else ExitStack():
            if signal_type == SignalType.DELETE:
                if is_meta_supported:
                    return model_cls.cqrs_delete(instance_data, meta=meta)

                return model_cls.cqrs_delete(instance_data)

            f_kw = {'previous_data': previous_data}
            if is_meta_supported:
                f_kw['meta'] = meta

            if signal_type == SignalType.SAVE:
                return model_cls.cqrs_save(instance_data, **f_kw)

            if signal_type == SignalType.SYNC:
                f_kw['sync'] = True
                return model_cls.cqrs_save(instance_data, **f_kw)

    except Error as e:
        pk_name = getattr(model_cls._meta.pk, 'name', 'id')
        pk_value = instance_data.get(pk_name)
        cqrs_revision = instance_data.get('cqrs_revision')

        logger.error(
            '{0}\nCQRS {1} error: pk = {2}, cqrs_revision = {3} ({4}).'.format(
                str(e), signal_type, pk_value, cqrs_revision, model_cls.CQRS_ID,
            ),
        )



#  Copyright © 2023 Ingram Micro Inc. All rights reserved.

import logging


logger = logging.getLogger('django-cqrs')


class LoggingMixin:
    _BASE_PAYLOAD_LOG_TEMPLATE = "CQRS is %s: pk = %s (%s), correlation_id = %s."

    @staticmethod
    def log_consumed(payload):
        """
        Args:
            payload (dj_cqrs.dataclasses.TransportPayload): Transport payload from master model.
        """
        msg = "CQRS is received: pk = %s (%s), correlation_id = %s."
        logger.info(msg, payload.pk, payload.cqrs_id, payload.correlation_id)

    @staticmethod
    def log_consumed_accepted(payload):
        """
        Args:
            payload (dj_cqrs.dataclasses.TransportPayload): Transport payload from master model.
        """
        msg = "CQRS is applied: pk = %s (%s), correlation_id = %s."
        logger.info(msg, payload.pk, payload.cqrs_id, payload.correlation_id)

    @staticmethod
    def log_consumed_denied(payload):
        """
        Args:
            payload (dj_cqrs.dataclasses.TransportPayload): Transport payload from master model.
        """
        msg = "CQRS is denied: pk = %s (%s), correlation_id = %s."
        logger.warning(msg, payload.pk, payload.cqrs_id, payload.correlation_id)

    @staticmethod
    def log_consumed_failed(payload):
        """
        Args:
            payload (dj_cqrs.dataclasses.TransportPayload): Transport payload from master model.
        """
        msg = (
            "CQRS is failed: pk = %s (%s), correlation_id = %s, retries = %s.",
        )
        logger.warning(
            msg, payload.pk, payload.cqrs_id, payload.correlation_id, payload.retries,
        )

    @staticmethod
    def log_dead_letter(payload):
        """
        Args:
            payload (dj_cqrs.dataclasses.TransportPayload): Transport payload from master model.
        """
        msg = "CQRS is added to dead letter queue: pk = %s (%s), correlation_id = %s."
        logger.warning(msg, payload.pk, payload.cqrs_id, payload.correlation_id)

    @staticmethod
    def log_delayed(payload, delay, eta):
        """
        Args:
            payload (dj_cqrs.dataclasses.TransportPayload): Transport payload from master model.
            delay (int): Seconds to wait before requeuing message.
            eta (datetime): Requeuing datetime.
        """
        msg = (
            "CQRS is delayed: pk = %s (%s), correlation_id = %s, delay = %s sec, eta = %s.",
        )
        logger.warning(
            msg, payload.pk, payload.cqrs_id, payload.correlation_id, delay, eta,
        )

    @staticmethod
    def log_requeued(payload):
        """
        Args:
            payload (dj_cqrs.dataclasses.TransportPayload): Transport payload from master model.
        """
        msg = (
            "CQRS is requeued: pk = %s (%s), correlation_id = %s.",
        )
        logger.warning(msg, payload.pk, payload.cqrs_id, payload.correlation_id)

    @staticmethod
    def log_produced(payload):
        """
        Args:
            payload (dj_cqrs.dataclasses.TransportPayload): Transport payload from master model.
        """
        msg = "CQRS is published: pk = %s (%s), correlation_id = %s."
        logger.info(msg, payload.pk, payload.cqrs_id, payload.correlation_id)



#  Copyright © 2023 Ingram Micro Inc. All rights reserved.

import sys

import ujson
from django.core.management.base import BaseCommand, CommandError

from dj_cqrs.registries import MasterRegistry


class Command(BaseCommand):
    help = 'Diff of deleted CQRS models pks from master diff stream.'

    @classmethod
    def serialize_out(cls, package):
        return ujson.dumps(package)

    @classmethod
    def deserialize_in(cls, package_line):
        return set(ujson.loads(package_line))

    def handle(self, *args, **options):
        with sys.stdin as f:
            first_line = f.readline()
            model = self._get_model(first_line)
            self.stdout.write(first_line.strip())

            for package_line in f:
                master_data = self.deserialize_in(package_line)

                exist_pks = set(
                    model.objects.filter(
                        pk__in=master_data,
                    ).values_list(
                        'pk', flat=True,
                    ),
                )
                diff_ids = list(master_data - exist_pks)
                if diff_ids:
                    self.stdout.write(self.serialize_out(diff_ids))
                    self.stderr.write('PK to delete: {0}'.format(str(diff_ids)))

    @staticmethod
    def _get_model(first_line):
        cqrs_id = first_line.split(',')[0]
        model = MasterRegistry.get_model_by_cqrs_id(cqrs_id)

        if not model:
            raise CommandError('Wrong CQRS ID: {0}!'.format(cqrs_id))

        return model
